;**********************************************************************
;                                                                     *
;    8 Channel PWM LED Chaser for PIC 16F628/A                        *
;                                                                     *
;    Filename:                  decoder.asm                           *
;    Date:                      04/03/2013                            *
;    File Version:              1.0.1                                 *
;                                                                     *
;    Author:                    Pete Griffiths                        *
;    Company:                   http://picprojects.org.uk             *
;                                                                     *
;                                                                     *
;**********************************************************************
;            PIC16F6287A/ 16F628A                                     *
;                                                                     *
;            A2 -| 1   18 |- A1                                       *
;            A3 -| 2   17 |- A0                                       *
;    lamp test  -| 3   16 |- invert                                   *
;         MCLR  -| 4   15 |- zero blank                               *
;          Vdd  -| 5   14 |- Vdd                                      *
;    en / latch -| 6   13 |- seg G                                    *
;         seg.A -| 7   12 |- seg.F                                    *
;         seg.B -| 8   11 |- seg.E                                    *
;         seg.C -| 9   10 |- seg.D                                    *
;                                                                     *
;                                                                     *
;   External pull-ups are needed on all inputs unless driven          *
;                                                                     *
;   lamp test active low                                              *
;   zero blank active low                                             *
;   en(able) low / latch high                                         *
;                                                                     *
;   Address inputs A0-A3                                              *
;                                                                     *
;   invert input high, outputs active high                            *
;   invert input low,  outputs active low                             *
;   invert input tested at power-on and mode set until next power-on  *
;                                                                     *
;   MCLR, reset input should be tied to Vdd via 1K resistor           *
;                                                                     *
;                                                                     *
;**********************************************************************
;                                                                     *
    processor 16F628

    #include <P16F628.INC>
  __CONFIG   _CP_OFF & _WDT_OFF & _PWRTE_ON & _INTRC_OSC_NOCLKOUT & _LVP_OFF & _MCLRE_ON

    errorlevel -302     ; suppress banksel warning messages during assembly


    ;------------------------------------------------------------------------------------
    ; define variables in General Purpose Register (GPR) memory
    ; See MPASM UG Section 4.8
    ; note: 16F84A GPR start at 0x0C but most other mid-range PICs start at 0x20
    ; We use 0x20 here for compatibilty with newer PICs

cblock 0x20
    V.seg.data: 32
    V.output
    V.address
    V.lamptest
    V.flags
endc




;--------------------------------------------------------------------------------------------------------------------
; Output port bit to 7-segment map

;-Segment data to I/O port bit map

seg.A  equ 1<<.1  ; PB1
seg.B  equ 1<<.2  ; PB2
seg.C  equ 1<<.3  ; PB3
seg.D  equ 1<<.4  ; PB4
seg.E  equ 1<<.5  ; PB5
seg.F  equ 1<<.6  ; PB6
seg.G  equ 1<<.7  ; PB7

;--------------------------------------------------------------------------------------------------------------------
; EEPROM data
; 4-bit Binary-to-Hex Decode
;
  org 0x2100

  ;   segA
  ; s      s
  ; e      e
  ; g      g
  ; F      B
  ;   segG
  ; s      s
  ; e      e
  ; g      g
  ; E      C
  ;   segD

  de  seg.A | seg.B | seg.C | seg.D | seg.E | seg.F  ; 0
  de  seg.B | seg.C                                  ; 1
  de  seg.A | seg.B | seg.D | seg.E | seg.G          ; 2
  de  seg.A | seg.B | seg.C | seg.D | seg.G          ; 3
  de  seg.B | seg.C | seg.F | seg.G                  ; 4

  de  seg.A | seg.C | seg.D | seg.F | seg.G          ; 5
  de  seg.A | seg.C | seg.D | seg.E | seg.F | seg.G  ; 6
  de  seg.A | seg.B | seg.C       ; 7
  de  seg.A | seg.B | seg.C | seg.D | seg.E | seg.F | seg.G  ; 8
  de  seg.A | seg.B | seg.C | seg.D | seg.F | seg.G  ; 9

  de  seg.A | seg.B | seg.C | seg.E | seg.F | seg.G  ; A
  de  seg.C | seg.D | seg.E | seg.F | seg.G          ; B
  de  seg.A | seg.D | seg.E | seg.F                  ; C
  de  seg.B | seg.C | seg.D | seg.E | seg.G          ; D
  de  seg.A | seg.D | seg.E | seg.F | seg.G          ; E
  de  seg.A | seg.E | seg.F | seg.G                  ; F

;--------------------------------------------------------------------------------------------------------------------
  ; embed firmware release info
  de " Binary to 7-segment decoder "
  de " V1.0.1 "
  de " 24/01/2015 "
  de " picprojects.org.uk "
;--------------------------------------------------------------------------------------------------------------------

;--- define port bits (PB)
PB.latch.en  equ  0  ; PortB, 0  Enable == 0 | Latch == 1

PB.A0        equ  0  ; PortA, 0
PB.A1        equ  1  ; PortA, 1
PB.A2        equ  2  ; PortA, 2
PB.A3        equ  3  ; PortA, 3
PB.lamp.test equ  4  ; PortA, 4  Lamp test input
PB.blank     equ  6  ; PortA, 6  Ripple blanking input
PB.invert    equ  7  ; PortA, 7  Invert output for common anode

;--- define flag bits
F.invert     equ  0  ; outputs are active low if flag bit set

;--------------------------------------------------------------------------------------------------------------------
; Program code starts here
;
RESET_VECTOR  org    0x000  ; see MPASM UG Section 4.50
              goto   START  ; START label is in pwmc_start.inc

INTERRUPT_V   org    0x004  ; Interrupt handler code must follow this origin statement

;-----------------------------------------------------------------------------
;
; Initialise peripherals and registers at startup
;
START
  movlw   0x07
  movwf   CMCON  ; disable comparator

  banksel TRISA  ; select register Bank 1

  movlw   1<<PB.invert | 1<<PB.lamp.test | 1<<PB.blank | 1<< PB.A3 | 1<< PB.A2 | 1<< PB.A1 | 1<< PB.A0
  movwf   TRISA  ;

  movlw   1<<PB.latch.en
  movwf   TRISB  ;


  ;--- all unused I/O lines on Port A need  to be tied to Vss

_setOptReg
  movlw  b'00000111'  ; setup option register
  ;        ||||||||---- PS0 - Timer 0: Prescalar 1:256
  ;        |||||||----- PS1
  ;        ||||||------ PS2
  ;        |||||------- PSA -  Assign prescaler to Timer0
  ;        ||||-------- TOSE - L toH  edge (don't care)
  ;        |||--------- TOCS - Timer0 uses IntClk
  ;        ||---------- INTEDG - falling edge RB0 (don't care)
  ;        |----------- RBPU - pull-ups enabled
  movwf   OPTION_REG

  banksel PORTA    ; select register Bank 0


; test invert input and set flag
; If input low, outputs will be active low, else active high
  clrf  V.flags
  ;btfss  PORTA, PB.invert
  btfsc PORTA, PB.invert  ; Changed to work with Evaluation PCB
  bsf   V.flags, F.invert

  ; read data from EEPROM table and store in GPR memory
  ; 0x20 to 0x2F holds segment data set with zero
  ; 0x30 to 0x3F holds segment data set with zero blanked
  ; When the zero blank input is high we read data from 0x20-0x2F
  ; When the zero blank input is low we read data from 0x30-0x3F

  clrf  V.address

read.next  movlw  V.seg.data
  movwf  FSR
  movfw  V.address
  addwf  FSR,F
 ;call   lookup
  call   E.read
  btfsc  V.flags, F.invert
  xorlw  0xFF

; copy data to tables starting at 0x20 and 0x30
  movwf  INDF
  bsf    FSR,4   ; point to table starting at 0x30 by setting this bit in FSR
  movwf  INDF
  incf   V.address,F
  btfss  V.address, 4  ; 2^4 = 16
  goto   read.next

; set the segment data for the zero to blank in the second table data set
  movlw  V.seg.data+.16
  movwf  FSR

; set up data for lamp test, all segments on
  movlw  0x00
  btfsc  V.flags, F.invert
  movlw  0xFF
  movwf  V.lamptest
  movwf  INDF

  movlw  V.seg.data
  movwf  FSR
;--------------------------------------------------------


main.loop
; if latch enable input high, data is latched, don't read port
  btfsc  PORTB, PB.latch.en  ; test latch enable input,
  goto   drive.seg

; if latch enable is low, we read value on PORTA and use it as index pointer
  movfw  PORTA      ; read portA
  andlw  0x0F       ; mask of bits so we keep only four low bits
  addlw  V.seg.data ; add to base address of segment data lookup table in GPR
  movwf  FSR        ; write address point to FSR index register

drive.seg
  bcf    FSR,4      ; clear bit 4 of FSR so we point to table starting at 0x20
  btfss  PORTA, PB.blank ; test zero blanking input
  bsf    FSR,4      ; if low, set bit 4 of FSR so we point to table starting at 0x30
  movfw  INDF       ; read segment data from table

  btfss  PORTA, PB.lamp.test  ; test lamp test input
  comf   V.lamptest,W  ; if input low, force all segment data active

  movwf  PORTB     ; write segment data to PORTB
  goto   main.loop


;-----------------------------------------------------------------------------
; EEPROM functions
;-----------------------------------------------------------------------------
; EEPROM Read
; Set EEPROM address to contents of W register
; Read EEPROM and return value in W register
E.read
  banksel EEADR
  movwf   EEADR

  bsf     EECON1, RD
  movfw   EEDATA
  banksel PORTA
  return

  end           ; end of instructions, always needed